"""
Dashboard widgets
"""

# Django
from django.contrib.auth.decorators import permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import render
from django.template.loader import render_to_string

# AA Forum
from aa_forum.helper.user import get_user_profile
from aa_forum.views.forum import _get_boards_with_unread_topics


def dashboard_widgets(request):
    """
    Return the unread topics

    :param request:
    :type request:
    :return:
    :rtype:
    """

    if request.user.has_perm("aa_forum.basic_access"):
        return render_to_string(
            template_name="aa_forum/view/widgets/dashboard-widgets.html",
            context={"user_profile": get_user_profile(user=request.user)},
            request=request,
        )

    return ""


@permission_required(perm="aa_forum.basic_access")
def ajax_unread_topics(request: WSGIRequest) -> HttpResponse:
    """
    AJAX Unread Topics widget

    :param request:
    :type request:
    :return:
    :rtype:
    """

    user_profile = get_user_profile(user=request.user)

    boards = _get_boards_with_unread_topics(request=request)

    if user_profile.show_unread_topics_dashboard_widget is True and boards.count() > 0:
        return render(
            template_name="aa_forum/partials/widgets/unread-topics.html",
            context={"boards": boards},
            request=request,
        )

    return HttpResponse(status=204)



"""
Tests for the models
"""

# Standard Library
import datetime as dt
from unittest.mock import patch

# Django
from django.contrib.auth.models import Group, User
from django.db.utils import IntegrityError
from django.test import TestCase
from django.urls import reverse
from django.utils.timezone import now

# AA Forum
from aa_forum.models import Board, Setting, Topic, get_sentinel_user
from aa_forum.tests.utils import (
    create_board,
    create_category,
    create_fake_messages,
    create_fake_user,
    create_last_message_seen,
    create_message,
    create_personal_message,
    create_setting,
    create_topic,
    my_get_setting,
)

MODELS_PATH = "aa_forum.models"


class TestGetSentinelUser(TestCase):
    """
    Tests for the get_sentinel_user function
    """

    def test_should_create_user_when_it_does_not_exist(self):
        """
        Test should create the sentinel user when it doesn't exist

        :return:
        :rtype:
        """

        # when
        user = get_sentinel_user()

        # then
        self.assertEqual(first=user.username, second="deleted")

    def test_should_return_user_when_it_does(self):
        """
        Test should return the sentinel user when it does exist

        :return:
        :rtype:
        """

        # given
        User.objects.create_user(username="deleted")

        # when
        user = get_sentinel_user()

        # then
        self.assertEqual(first=user.username, second="deleted")


class TestBoard(TestCase):
    """
    Tests for Board
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users

        :return:
        :rtype:
        """

        super().setUpClass()

        cls.user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        cls.group = Group.objects.create(name="Superhero")

    def setUp(self) -> None:
        """
        Set up category

        :return:
        :rtype:
        """

        self.category = create_category(name="Science")

    def test_model_string_names(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        board = create_board(name="Physics", category=self.category)

        self.assertEqual(first=str(board), second="Physics")

    def test_should_update_last_message_when_message_created(self):
        """
        Test should update last message when message created

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        topic = create_topic(board=board)

        # when
        message = create_message(topic=topic, user_created=self.user)

        # then
        board.refresh_from_db()
        self.assertEqual(first=board.last_message, second=message)

    def test_should_update_last_message_when_message_created_in_child_board(self):
        """
        Test should update last message when message created in child board

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        child_board = create_board(category=self.category, parent_board=board)
        topic_board = create_topic(board=board)
        topic_child_board = create_topic(board=child_board)
        create_message(topic=topic_board, user_created=self.user)

        # when
        message_child_board = create_message(
            topic=topic_child_board, user_created=self.user
        )

        # then
        board.refresh_from_db()
        self.assertEqual(first=board.last_message, second=message_child_board)

    def test_should_return_url(self):
        """
        Test should return board URL

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category, name="Physics")

        # when
        url = board.get_absolute_url()

        # then
        self.assertURLEqual(
            url1=url,
            url2=reverse(viewname="aa_forum:forum_board", args=["science", "physics"]),
        )

    def test_should_return_board_with_no_groups(self):
        """
        Test should return board with no groups

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)

        # when
        result = Board.objects.user_has_access(self.user).get(pk=board.pk)

        # then
        self.assertTrue(expr=result)

    def test_should_return_board_for_group_member(self):
        """
        Test should return board for group member

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)

        self.user.groups.add(self.group)

        # when
        result = Board.objects.user_has_access(self.user).get(pk=board.pk)

        # then
        self.assertTrue(expr=result)

    def test_should_return_child_board_for_group_member(self):
        """
        Test should return restricted child board for group member

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)

        self.user.groups.add(self.group)

        child_board = create_board(category=self.category, parent_board=board)

        # when
        result = Board.objects.user_has_access(user=self.user).get(pk=child_board.pk)

        # then
        self.assertTrue(expr=result)

    def test_should_not_return_board_for_non_group_member(self):
        """
        Test should not return restricted board for non group member

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)

        with self.assertRaises(expected_exception=Board.DoesNotExist):
            Board.objects.user_has_access(user=self.user).get(pk=board.pk)

    def test_should_not_return_child_board_for_non_group_member(self):
        """
        Test should not return restricted child board for non group member

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        board.groups.add(self.group)
        board_2 = create_board(category=self.category, parent_board=board)

        with self.assertRaises(expected_exception=Board.DoesNotExist):
            Board.objects.user_has_access(user=self.user).get(pk=board_2.pk)

    def test_should_generate_new_slug_when_slug_already_exists(self):
        """
        Test should generate a new slug when board-slug already exists

        :return:
        :rtype:
        """

        # given
        board = create_board(category=self.category)
        board.slug = "dummy"  # we are faking the same slug here
        board.save()

        # when
        board = create_board(name="Dummy", category=self.category)

        # then
        self.assertEqual(first=board.slug, second="dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        """
        Test should not allow creation with hyphen slugs (reserved for admin pages)

        :return:
        :rtype:
        """

        # when
        board = create_board(name="Dummy", category=self.category, slug="-")

        # then
        self.assertEqual(first=board.slug, second="dummy")

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_should_create_slug_with_name_matching_internal_url_prefix(self):
        """
        Test should create slug with name matching internal URL prefix

        :return:
        :rtype:
        """

        # when
        category = create_board(name="-", category=self.category)

        # then
        self.assertEqual(first=category.slug, second="hyphen")

    def test_should_handle_empty_board_slug(self):
        """
        Handling an empty slug returned by Board.save() when the board
        name is just special characters, which are getting stripped by slug generation

        :return:
        :rtype:
        """

        # when
        board = create_board(name="@#$%", category=self.category)

        # then
        expected_slug = f"{board.__class__.__name__}-{board.pk}".lower()
        self.assertEqual(first=board.slug, second=expected_slug)

    def test_should_translate_nonlatin_letters_in_slug(self):
        """
        Test that non latin letters in a slug are translated

        :return:
        :rtype:
        """

        board = create_board(name="дрифтерке, рорки в док", category=self.category)

        expected_slug = "drifterke-rorki-v-dok"

        self.assertEqual(first=board.slug, second=expected_slug)

    def test_should_create_topic_in_board(self):
        """
        Test that a new topic is created in a given board

        :return:
        :rtype:
        """

        board = create_board(name="Test Board", category=self.category)
        topic = board.new_topic(subject="Foobar", message="Foobar", user=self.user)

        topic_last_created = Topic.objects.last()

        self.assertEqual(first=topic.subject, second=topic_last_created.subject)
        self.assertEqual(first=topic.board_id, second=topic_last_created.board_id)

    def test_should_raise_topic_already_exists_exception(self):
        """
        Test that the Board.TopicAlreadyExists exception is raised

        :return:
        :rtype:
        """

        board = create_board(name="Test Board", category=self.category)
        existing_topic = board.new_topic(
            subject="Foobar", message="Foobar", user=self.user
        )

        expected_exception = Board.TopicAlreadyExists

        existing_topic_url = reverse(
            viewname="aa_forum:forum_topic",
            kwargs={
                "category_slug": board.category.slug,
                "board_slug": board.slug,
                "topic_slug": existing_topic.slug,
            },
        )

        expected_message = f'<h4>Warning!</h4><p>There is already a topic with the exact same subject in this board.</p><p>See here: <a href="{existing_topic_url}">{existing_topic.subject}</a></p>'  # pylint: disable=line-too-long

        with self.assertRaises(expected_exception=expected_exception):
            board.new_topic(subject="Foobar", message="Foobar", user=self.user)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            board.new_topic(subject="Foobar", message="Foobar", user=self.user)


class TestCategory(TestCase):
    """
    Tests for Category
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.user = create_fake_user(character_id=1001, character_name="Bruce Wayne")

    def test_model_string_names(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        category = create_category(name="Science")

        self.assertEqual(first=str(category), second="Science")

    def test_should_create_new_category_with_slug(self):
        """
        Test should create a new category with slug

        :return:
        :rtype:
        """

        # when
        category = create_category(name="Science")

        # then
        self.assertEqual(first=category.slug, second="science")

    def test_should_keep_existing_slug_when_changing_name(self):
        """
        Test should keep existing slug when changing name

        :return:
        :rtype:
        """

        category = create_category(name="Science")

        # when
        category.name = "Politics"
        category.save()

        # then
        category.refresh_from_db()
        self.assertEqual(first=category.slug, second="science")

    def test_should_generate_new_slug_when_slug_already_exists(self):
        """
        Test should generate a new slug when slug already exists

        :return:
        :rtype:
        """

        # given
        category = create_category(name="Science")
        category.slug = "dummy"  # we are faking the same slug here
        category.save()

        # when
        category = create_category(name="Dummy")

        # then
        self.assertEqual(first=category.slug, second="dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        """
        Test should not allow creation with hyphen slugs (reserved for admin pages)

        :return:
        :rtype:
        """

        # when
        category = create_category(name="Science", slug="-")

        # then
        self.assertEqual(first=category.slug, second="science")

    def test_should_not_allow_to_update_slug_to_hyphen(self):
        """
        Test should not allow updating slug to hyphen (reserved for admin pages)

        :return:
        :rtype:
        """

        # given
        category = create_category(name="Science")

        # when
        category.slug = "-"
        category.save()

        # then
        category.refresh_from_db()
        self.assertEqual(first=category.slug, second="science-1")

    def test_should_translate_nonlatin_letters_in_slug(self):
        """
        Test that non latin letters in a slug are translated

        :return:
        :rtype:
        """

        category = create_category(name="дрифтерке, рорки в док")

        expected_slug = "drifterke-rorki-v-dok"

        self.assertEqual(first=category.slug, second=expected_slug)

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_should_create_new_category_with_name_matching_internal_url_prefix(self):
        """
        Test should create a new category with name matching internal URL prefix

        :return:
        :rtype:
        """

        # when
        category = create_category(name="-")

        # then
        self.assertEqual(first=category.slug, second="hyphen")

    def test_should_handle_empty_category_slug(self):
        """
        Handling an empty slug returned by Category.save() when the category
        name is just special characters, which are getting stripped by slug generation

        :return:
        :rtype:
        """

        # when
        category = create_category(name="@#$%")

        # then
        expected_slug = f"{category.__class__.__name__}-{category.pk}".lower()
        self.assertEqual(first=category.slug, second=expected_slug)


@patch(MODELS_PATH + ".Setting.objects.get_setting", new=my_get_setting)
class TestMessage(TestCase):
    """
    Tests for Message
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup

        :return:
        :rtype:
        """
        super().setUpClass()

        cls.user = create_fake_user(character_id=1001, character_name="Bruce Wayne")

    def setUp(self) -> None:
        """
        Setup

        :return:
        :rtype:
        """

        category = create_category()

        self.board = create_board(category=category)
        self.child_board = create_board(category=category, parent_board=self.board)
        self.topic = create_topic(board=self.board)
        self.child_topic = create_topic(board=self.child_board)

    def test_model_string_names(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        # when
        message = create_message(topic=self.topic, user_created=self.user)

        # then
        self.assertEqual(first=str(message), second=str(message.pk))

    def test_should_update_first_and_last_messages_when_saving_1(self):
        """
        Test should update the first and last message when saving

        :return:
        :rtype:
        """

        # when
        message = create_message(topic=self.topic, user_created=self.user)

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.topic.last_message, second=message)
        self.assertEqual(first=self.topic.first_message, second=message)
        self.assertEqual(first=self.board.last_message, second=message)
        self.assertEqual(first=self.board.first_message, second=message)

    def test_should_update_first_and_last_messages_when_saving_2(self):
        """
        Test should update the first and last message when saving multiple messages

        :return:
        :rtype:
        """

        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)

        # when
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.topic.last_message, second=message_2)
        self.assertEqual(first=self.topic.first_message, second=message_1)
        self.assertEqual(first=self.board.last_message, second=message_2)
        self.assertEqual(first=self.board.first_message, second=message_1)

    def test_should_update_first_and_last_messages_when_saving_3(self):
        """
        Test should update the first and last message when saving

        :return:
        :rtype:
        """

        # when
        message = create_message(topic=self.child_topic, user_created=self.user)

        # then
        self.child_topic.refresh_from_db()
        self.child_board.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.child_topic.last_message, second=message)
        self.assertEqual(first=self.child_topic.first_message, second=message)
        self.assertEqual(first=self.child_board.last_message, second=message)
        self.assertEqual(first=self.child_board.first_message, second=message)
        self.assertEqual(first=self.board.last_message, second=message)
        self.assertEqual(first=self.board.first_message, second=message)

    def test_should_update_first_and_last_messages_when_saving_4(self):
        """
        Test should update the first and last message when saving

        :return:
        :rtype:
        """

        # given
        message = create_message(topic=self.topic, user_created=self.user)

        # when
        child_message = create_message(topic=self.child_topic, user_created=self.user)

        # then
        self.child_topic.refresh_from_db()
        self.child_board.refresh_from_db()
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.child_topic.last_message, second=child_message)
        self.assertEqual(first=self.child_topic.first_message, second=child_message)
        self.assertEqual(first=self.child_board.last_message, second=child_message)
        self.assertEqual(first=self.child_board.first_message, second=child_message)
        self.assertEqual(first=self.topic.last_message, second=message)
        self.assertEqual(first=self.topic.last_message, second=message)
        self.assertEqual(first=self.board.last_message, second=child_message)
        self.assertEqual(first=self.board.first_message, second=child_message)

    def test_should_update_first_and_last_messages_when_saving_5(self):
        """
        Test should update the first and last message when saving

        :return:
        :rtype:
        """

        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # when
        child_message_1 = create_message(topic=self.child_topic, user_created=self.user)
        child_message_2 = create_message(topic=self.child_topic, user_created=self.user)

        # then
        self.child_topic.refresh_from_db()
        self.child_board.refresh_from_db()
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.child_topic.first_message, second=child_message_1)
        self.assertEqual(first=self.child_topic.last_message, second=child_message_2)
        self.assertEqual(first=self.child_board.first_message, second=child_message_1)
        self.assertEqual(first=self.child_board.last_message, second=child_message_2)
        self.assertEqual(first=self.topic.first_message, second=message_1)
        self.assertEqual(first=self.topic.last_message, second=message_2)
        self.assertEqual(first=self.board.first_message, second=child_message_1)
        self.assertEqual(first=self.board.last_message, second=child_message_2)

    def test_should_update_message_references_when_deleting_last_message_1(self):
        """
        Test should updat thee first and last message when deleting the last message

        :return:
        :rtype:
        """

        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # when
        message_2.delete()

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.topic.last_message, second=message_1)
        self.assertEqual(first=self.topic.first_message, second=message_1)
        self.assertEqual(first=self.board.last_message, second=message_1)
        self.assertEqual(first=self.board.first_message, second=message_1)

    def test_should_update_message_references_when_deleting_last_message_2(self):
        """
        Test should update the first and last message when deleting the last message

        :return:
        :rtype:
        """

        # given
        message_1 = create_message(topic=self.topic, user_created=self.user)
        message_2 = create_message(topic=self.topic, user_created=self.user)

        # when
        message_1.delete()

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.topic.last_message, second=message_2)
        self.assertEqual(first=self.topic.first_message, second=message_2)
        self.assertEqual(first=self.board.last_message, second=message_2)
        self.assertEqual(first=self.board.first_message, second=message_2)

    def test_should_update_message_references_when_deleting_last_message_3(self):
        """
        Test should update the first and last message when deleting the last message

        :return:
        :rtype:
        """

        # given
        message = create_message(topic=self.topic, user_created=self.user)
        child_message = create_message(topic=self.child_topic, user_created=self.user)

        # when
        child_message.delete()

        # then
        self.topic.refresh_from_db()
        self.child_topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertEqual(first=self.topic.last_message, second=message)
        self.assertEqual(first=self.topic.first_message, second=message)
        self.assertIsNone(obj=self.child_topic.last_message)
        self.assertIsNone(obj=self.child_topic.first_message)
        self.assertEqual(first=self.board.last_message, second=message)
        self.assertEqual(first=self.board.first_message, second=message)

    def test_should_update_message_references_when_deleting_last_message_4(self):
        """
        Test should update the first and last message when deleting the last message

        :return:
        :rtype:
        """

        # given
        message = create_message(topic=self.topic, user_created=self.user)
        child_message = create_message(topic=self.child_topic, user_created=self.user)

        # when
        message.delete()

        # then
        self.topic.refresh_from_db()
        self.child_topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertIsNone(obj=self.topic.last_message)
        self.assertIsNone(obj=self.topic.first_message)
        self.assertEqual(first=self.child_topic.last_message, second=child_message)
        self.assertEqual(first=self.child_topic.first_message, second=child_message)
        self.assertEqual(first=self.board.last_message, second=child_message)
        self.assertEqual(first=self.board.first_message, second=child_message)

    def test_should_reset_message_references_when_deleting_last_message_1(self):
        """
        Test should reset message references when deleting the last message

        :return:
        :rtype:
        """

        # given
        message = create_message(topic=self.topic, user_created=self.user)

        # when
        message.delete()

        # then
        self.topic.refresh_from_db()
        self.board.refresh_from_db()
        self.assertIsNone(obj=self.topic.last_message)
        self.assertIsNone(obj=self.topic.first_message)
        self.assertIsNone(obj=self.board.last_message)
        self.assertIsNone(obj=self.board.first_message)

    def test_should_return_url_first_page(self):
        """
        Test should return URL to the first page

        :return:
        :rtype:
        """

        # given
        message = create_message(topic=self.topic, user_created=self.user)
        url = message.get_absolute_url()

        # then
        self.assertURLEqual(
            url1=url,
            url2=reverse(
                viewname="aa_forum:forum_topic",
                args=(
                    self.topic.board.category.slug,
                    self.topic.board.slug,
                    self.topic.slug,
                ),
            )
            + f"#message-{message.pk}",
        )

    def test_should_return_url_other_pages(self):
        """
        Test should return URL to other pages

        :return:
        :rtype:
        """

        # given
        create_fake_messages(topic=self.topic, amount=9)
        message = self.topic.messages.order_by("time_posted").last()
        url = message.get_absolute_url()

        # then
        self.assertURLEqual(
            url1=url,
            url2=reverse(
                viewname="aa_forum:forum_topic",
                args=(
                    self.topic.board.category.slug,
                    self.topic.board.slug,
                    self.topic.slug,
                    2,
                ),
            )
            + f"#message-{message.pk}",
        )


class TestTopic(TestCase):
    """
    Tests for Topic
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup

        :return:
        :rtype:
        """

        super().setUpClass()

        cls.user = create_fake_user(character_id=1001, character_name="Bruce Wayne")

    def setUp(self) -> None:
        """
        Setup

        :return:
        :rtype:
        """

        category = create_category(name="Science")

        self.board = create_board(category=category, name="Physics")
        self.child_board = create_board(category=category, parent_board=self.board)

    def test_model_string_names(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        topic = create_topic(subject="Mysteries", board=self.board)

        self.assertEqual(first=str(topic), second="Mysteries")

    def test_should_update_last_message_when_message_is_created(self):
        """
        Test should update the last message when a message is created

        :return:
        :rtype:
        """

        # given
        topic = create_topic(subject="Mysteries", board=self.board)

        # when
        message = create_message(topic=topic, user_created=self.user)

        # then
        topic.refresh_from_db()
        self.assertEqual(first=topic.last_message, second=message)

    def test_should_update_last_message_when_message_is_created_child_board(self):
        """
        Test should update the last message when a message is created in child board

        :return:
        :rtype:
        """

        # given
        child_topic = create_topic(board=self.child_board)

        # when
        child_message = create_message(topic=child_topic, user_created=self.user)

        # then
        child_topic.refresh_from_db()
        self.assertEqual(first=child_topic.last_message, second=child_message)

    def test_should_update_last_message_after_topic_deletion(self):
        """
        Test should update last message when the topic is deleted

        :return:
        :rtype:
        """

        # given
        topic_1 = create_topic(board=self.board)
        topic_2 = create_topic(board=self.board)
        my_now = now() - dt.timedelta(hours=1)

        with patch("django.utils.timezone.now", lambda: my_now):
            message_1 = create_message(topic=topic_1, user_created=self.user)

        message_2 = create_message(topic=topic_2, user_created=self.user)

        self.board.refresh_from_db()
        self.assertEqual(first=self.board.last_message, second=message_2)

        # when
        topic_2.delete()

        # then
        self.board.refresh_from_db()
        self.assertEqual(first=self.board.last_message, second=message_1)

    def test_should_not_update_last_message_after_topic_deletion(self):
        """
        Test should not update the last message when the topic is deleted

        :return:
        :rtype:
        """

        # given
        topic_1 = create_topic(board=self.board)
        topic_2 = create_topic(board=self.board)
        my_now = now() - dt.timedelta(hours=1)

        with patch("django.utils.timezone.now", lambda: my_now):
            create_message(topic=topic_1, user_created=self.user)

        message_2 = create_message(topic=topic_2, user_created=self.user)

        self.board.refresh_from_db()
        self.assertEqual(first=self.board.last_message, second=message_2)

        # when
        topic_1.delete()

        # then
        self.board.refresh_from_db()
        self.assertEqual(first=self.board.last_message, second=message_2)

    def test_should_return_url(self):
        """
        Test should return topic URL

        :return:
        :rtype:
        """

        # given
        topic = create_topic(subject="Mysteries", board=self.board)

        # when
        url = topic.get_absolute_url()

        # then
        self.assertURLEqual(
            url1=url,
            url2=reverse(
                viewname="aa_forum:forum_topic",
                args=["science", "physics", "mysteries"],
            ),
        )

    def test_should_generate_new_slug_when_slug_already_exists(self):
        """
        Test should generate a new slug when slug already exists

        :return:
        :rtype:
        """

        # given
        topic = create_topic(board=self.board)
        topic.slug = "dummy"  # we are faking the same slug here
        topic.save()

        # when
        topic = create_topic(subject="Dummy", board=self.board)

        # then
        self.assertEqual(first=topic.slug, second="dummy-1")

    def test_should_not_allow_creation_with_hyphen_slugs(self):
        """
        Test should not allow creation with hyphen slugs

        :return:
        :rtype:
        """

        # when
        topic = create_topic(subject="Dummy", board=self.board, slug="-")

        # then
        self.assertEqual(first=topic.slug, second="dummy")

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_should_create_slug_with_name_matching_internal_url_prefix(self):
        """
        Test should create slug with name matching internal URL prefix

        :return:
        :rtype:
        """

        # when
        topic = create_topic(subject="-", board=self.board)

        # then
        self.assertEqual(first=topic.slug, second="hyphen")

    def test_should_handle_empty_topic_slug(self):
        """
        Handling an empty slug returned by Topic.save() when the topic
        name is just special characters, which are getting stripped by slug generation

        :return:
        :rtype:
        """

        # when
        topic = create_topic(subject="@#$%", board=self.board)

        # then
        expected_slug = f"{topic.__class__.__name__}-{topic.pk}".lower()
        self.assertEqual(first=topic.slug, second=expected_slug)

    def test_should_translate_nonlatin_letters_in_slug(self):
        """
        Test that non latin letters in a slug are translated to latin

        :return:
        :rtype:
        """

        topic = create_topic(subject="дрифтерке, рорки в док", board=self.board)

        expected_slug = "drifterke-rorki-v-dok"

        self.assertEqual(first=topic.slug, second=expected_slug)


class TestPersonalMessage(TestCase):
    """
    Tests for PersonalMessage
    """

    def test_str(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        # with
        message = create_personal_message(subject="Subject")

        # then
        self.assertEqual(
            first=str(message), second='"Subject" from Bruce_Wayne to Lex_Luthor'
        )

    @patch(MODELS_PATH + ".INTERNAL_URL_PREFIX", "-")
    def test_can_create_personal_message(self):
        """
        Test can create a personal message

        :return:
        :rtype:
        """

        # with
        message = create_personal_message(subject="subject")

        # then
        self.assertEqual(first=message.subject, second="subject")


class LastMessageSeen(TestCase):
    """
    Tests for LastMessageSeen
    """

    def test_str(self):
        """
        Test model string names

        :return:
        :rtype:
        """

        # given
        topic = create_topic(subject="Alpha")
        user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        message_time = dt.datetime(year=2022, month=1, day=12, hour=17, minute=30)
        message = create_last_message_seen(
            topic=topic, user=user, message_time=message_time
        )

        # when/then
        result = str(message)

        # then
        self.assertEqual(first=result, second="Alpha-Bruce_Wayne-2022-01-12 17:30:00")


class TestSetting(TestCase):
    """
    Tests for Setting
    """

    def test_model_string_name(self):
        """
        Test model string name

        :return:
        :rtype:
        """

        # given
        setting = Setting.objects.get(pk=1)

        # when/then
        self.assertEqual(first=str(setting), second="Forum settings")

    def test_setting_save(self):
        """
        Test if there can't be another setting created
        and the existing setting is changed instead

        :return:
        :rtype:
        """

        # given
        messages_per_page = 25
        topics_per_page = 25
        setting = Setting(
            pk=2, messages_per_page=messages_per_page, topics_per_page=topics_per_page
        )
        setting.save()

        # then
        self.assertEqual(first=setting.pk, second=1)
        self.assertEqual(first=setting.messages_per_page, second=messages_per_page)
        self.assertEqual(first=setting.topics_per_page, second=topics_per_page)

    def test_setting_create(self):
        """
        Test that create method throwing the following exception
        django.db.utils.IntegrityError: (1062, "Duplicate entry '1' for key 'PRIMARY'")

        :return:
        :rtype:
        """

        # No pk given
        # with self.assertRaisesMessage(
        #     IntegrityError, "(1062, \"Duplicate entry '1' for key 'PRIMARY'\")"
        # ):
        with self.assertRaises(expected_exception=IntegrityError):
            create_setting()

    def test_setting_create_with_pk(self):
        """
        Test that create method throwing the following exception no matter the given pk
        django.db.utils.IntegrityError: (1062, "Duplicate entry '1' for key 'PRIMARY'")

        :return:
        :rtype:
        """

        # Set pk=2
        # with self.assertRaisesMessage(
        #     IntegrityError, "(1062, \"Duplicate entry '1' for key 'PRIMARY'\")"
        # ):
        with self.assertRaises(expected_exception=IntegrityError):
            create_setting(pk=2)

    def test_cannot_be_deleted(self):
        """
        Test that the settings object cannot be deleted

        :return:
        :rtype:
        """

        # given
        settings_old = Setting.objects.get(pk=1)

        # when
        Setting.objects.all().delete()

        # then
        settings = Setting.objects.all()
        settings_first = settings.first()

        # See that there is still only ONE Setting object
        self.assertEqual(first=settings.count(), second=1)

        # Check that both of our objects are identical
        self.assertEqual(first=settings_old, second=settings_first)



"""
Our signals
"""

# Django
from django.db.models.signals import post_save
from django.dispatch import receiver

# AA Forum
from aa_forum.models import Board


@receiver(post_save, sender=Board)
def sync_parent_board_access_to_child_board(
    sender,  # pylint: disable=unused-argument
    instance,
    **kwargs,  # pylint: disable=unused-argument
):
    """
    Sync parent board access to child board

    :param sender:
    :type sender:
    :param instance:
    :type instance:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:
    """

    if instance.parent_board:
        parent_board = Board.objects.get(pk=instance.parent_board.pk)

        instance.groups.set(parent_board.groups.all())
    else:
        child_boards = instance.child_boards.all()

        for child_board in child_boards:
            child_board.groups.set(instance.groups.all())



"""
AA Forum Admin
"""

# Django
from django.contrib import admin
from django.utils.safestring import mark_safe

# AA Forum
from aa_forum.models import Board, Category, Setting, Topic, UserProfile


class SingletonModelAdmin(admin.ModelAdmin):
    """
    Singleton Model Admin
    Prevents Django admin users deleting the singleton or adding extra rows.
    """

    actions = None  # Removes the default delete action.

    def has_add_permission(self, request):  # pylint: disable=unused-argument
        """
        Has "add" permissions

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return self.model.objects.all().count() == 0

    def has_change_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "change" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return True

    def has_delete_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "delete" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False


class BaseReadOnlyAdminMixin:
    """
    Base "Read-Only" mixin for admin models
    """

    def has_add_permission(self, request):  # pylint: disable=unused-argument
        """
        Has add permissions

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return False

    def has_change_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "change" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False

    def has_delete_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "delete" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False


@admin.register(Category)
class CategoryAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Category admin
    """

    list_display = ("name", "slug", "_board_count")
    exclude = ("is_collapsible",)

    def _board_count(self, obj):
        """
        Return the board count per category

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.boards.count()


@admin.register(Board)
class BoardAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Board admin
    """

    list_display = (
        "name",
        "slug",
        "parent_board",
        "_groups",
        "category",
        "_topics_count",
    )

    def _groups(self, obj):
        """
        Return the groups per board

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        groups = obj.groups.all()

        if groups.count() > 0:
            return mark_safe("<br>".join([group.name for group in groups]))

        return ""

    def _topics_count(self, obj):
        """
        Return the topic count per board

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.topics.count()


@admin.register(Topic)
class TopicAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Topic admin
    """

    list_display = ("subject", "slug", "board", "_messages_count")

    def _messages_count(self, obj):
        """
        Return the message count per topic

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return obj.messages.count()


@admin.register(Setting)
class SettingAdmin(SingletonModelAdmin):
    """
    Setting Admin
    """


@admin.register(UserProfile)
class UserProfileAdmin(admin.ModelAdmin):
    """
    User Profile Admin
    """

    list_display = ("user",)
    readonly_fields = ("user",)



# Generated by Django 3.2.7 on 2021-09-10 13:37

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aa_forum", "0002_default_settings"),
    ]

    operations = [
        migrations.AddField(
            model_name="board",
            name="discord_webhook",
            field=models.CharField(blank=True, default=None, max_length=254, null=True),
        ),
    ]



# Generated by Django 4.0.3 on 2022-04-08 18:44

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("aa_forum", "0004_board_use_webhook_for_replies"),
    ]

    operations = [
        migrations.AddField(
            model_name="board",
            name="announcement_groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="User in at least one of the selected groups will be able to start topics in 'Announcement Boards'. If no group is selected, only forum admins can do so.",
                related_name="aa_forum_boards_group_start_topic_restriction",
                to="auth.group",
            ),
        ),
        migrations.AddField(
            model_name="board",
            name="is_announcement_board",
            field=models.BooleanField(
                default=False,
                help_text="Mark this board as an 'Announcement Board', meaning that only certain selected groups can start new topics. All others who have access to this board will still be able to discuss in the topics though. (Default: NO)",
            ),
        ),
        migrations.AlterField(
            model_name="board",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="User in at least one of these groups have access to this board. If no groups are selected, everyone with access to the forum has also access to this board.",
                related_name="aa_forum_boards_group_restriction",
                to="auth.group",
            ),
        ),
    ]



"""
App config
"""

# Django
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

# AA Forum
from aa_forum import __version__


class AaForumConfig(AppConfig):
    """
    AA Forum App Config
    """

    name = "aa_forum"
    label = "aa_forum"
    verbose_name = _(f"AA Forum v{__version__}")

    def ready(self):
        """
        Make sure we can utilize signals

        :return:
        :rtype:
        """

        # AA Forum
        import aa_forum.signals  # noqa: F401 # pylint: disable=unused-import, import-outside-toplevel



# Django
from django.db import migrations


def on_migrate(apps, schema_editor):
    """
    Just pass through this one
    :param apps:
    :param schema_editor:
    :return:
    """

    pass


def on_migrate_zero(apps, schema_editor):
    """
    Remove all personal messages before jumping back to 0012
    :param apps:
    :param schema_editor:
    :return:
    """

    PersonalMessage = apps.get_model("aa_forum", "PersonalMessage")
    db_alias = schema_editor.connection.alias
    PersonalMessage.objects.using(db_alias).delete()


class Migration(migrations.Migration):
    dependencies = [
        ("aa_forum", "0012_personal_messages"),
    ]

    operations = [migrations.RunPython(on_migrate, on_migrate_zero)]



"""
Tests for the template tags
"""

# Standard Library
from datetime import datetime

# Third Party
from dateutil import parser

# Django
from django.template import TemplateSyntaxError
from django.test import TestCase, modify_settings

# Alliance Auth
from allianceauth.tests.auth_utils import AuthUtils

# Alliance Auth (External Libs)
from app_utils.urls import reverse as reverse_url

# AA Forum
from aa_forum import __version__
from aa_forum.models import PersonalMessage, get_sentinel_user
from aa_forum.templatetags.aa_forum import personal_message_unread_count
from aa_forum.tests.utils import create_fake_user, render_template


class TestMainCharacterName(TestCase):
    """
    Tests for main_character_name template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up template

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.template = "{% load aa_forum %}{{ user|aa_forum_main_character_name }}"

    def test_should_contain_character_name_for_users_with_main(self):
        """
        Test should contain a character name for user with a main set

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="Bruce Wayne")

    def test_should_contain_user_character_name_for_users_without_main(self):
        """
        Should return username for users without a main character

        :return:
        :rtype:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="john")

    def test_should_return_deleted_user_for_sentinel_user(self):
        """
        Should return "deleted" for sentinel user

        :return:
        :rtype:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="deleted")

    def test_should_be_empty_for_none(self):
        """
        Test should be empty

        :return:
        :rtype:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")


class TestMainCharacterId(TestCase):
    """
    Tests for aa_forum_main_character_id template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up template

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.template = "{% load aa_forum %}{{ user|aa_forum_main_character_id }}"

    def test_should_contain_character_id_for_users_with_main(self):
        """
        Test should contain main character ID for users with main

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(character_id=1001, character_name="Bruce Wayne")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1001")

    def test_should_contain_dummy_id_for_users_without_main(self):
        """
        Test should contain dummy ID (1) for users without main character

        :return:
        :rtype:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")

    def test_should_be_dummy_id_for_sentinel_user(self):
        """
        Test should return dummy ID (1) for sentinel user

        :return:
        :rtype:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")

    def test_should_be_dummy_id_for_none(self):
        """
        Test should return dummy ID (1) for None

        :return:
        :rtype:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")


class TestMainCharacterCorporationName(TestCase):
    """
    Tests for aa_forum_main_character_corporation_name template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up template

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.template = (
            "{% load aa_forum %}{{ user|aa_forum_main_character_corporation_name }}"
        )

    def test_should_contain_corp_name_for_users_with_main(self):
        """
        Test should return corporation name for users with a main character

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
        )
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="Wayne Tech Inc.")

    def test_should_be_empty_for_users_without_main(self):
        """
        Test should be empty for users without a main character

        :return:
        :rtype:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")

    def test_should_be_empty_for_sentinel_user(self):
        """
        Test should be empty for sentinel user

        :return:
        :rtype:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")

    def test_should_be_empty_for_none(self):
        """
        Test should be empty for None

        :return:
        :rtype:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")


class TestMainCorporationId(TestCase):
    """
    Tests for aa_forum_main_character_corporation_id template tag

    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up template

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.template = (
            "{% load aa_forum %}{{ user|aa_forum_main_character_corporation_id }}"
        )

    def test_should_contain_corporation_id_for_users_with_main(self):
        """
        Test should return the main character corp ID for users with a main character

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
        )
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="2001")

    def test_should_be_dummy_id_for_users_without_main(self):
        """
        Test should return dummy ID (1) for users without a main character

        :return:
        :rtype:
        """

        # given
        user = AuthUtils.create_user(username="john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)
        # then

        self.assertEqual(first=result, second="1")

    def test_should_be_dummy_id_for_sentinel_user(self):
        """
        Test should return dummy ID (1) for sentinel user

        :return:
        :rtype:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")

    def test_should_be_dummy_id_for_none(self):
        """
        Test should return dummy ID (1) for None

        :return:
        :rtype:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")


class TestMainCharacterAllianceName(TestCase):
    """
    Tests for aa_forum_main_character_alliance_name template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up template

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.template = (
            "{% load aa_forum %}{{ user|aa_forum_main_character_alliance_name }}"
        )

    def test_should_contain_alliance_name_for_users_with_main(self):
        """
        Test should return the main character alliance name for users with a main character

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )

        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="Wayne Enterprices")

    def test_should_be_empty_for_users_without_main(self):
        """
        Test should be empty for users without a main character

        :return:
        :rtype:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")

    def test_should_be_empty_when_main_is_not_in_an_alliance(self):
        """
        Test should be empty when the main character is not in an alliance

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(
            character_id=2012,
            character_name="William Riker",
            corporation_id=2012,
            corporation_name="Starfleet",
            corporation_ticker="SF",
            alliance_id=None,
        )

        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")

    def test_should_be_empty_for_sentinel_user(self):
        """
        Test should be empty for sentinel user

        :return:
        :rtype:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")

    def test_should_be_empty_for_none(self):
        """
        Test should be empty for None

        :return:
        :rtype:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="")


class TestMainAllianceId(TestCase):
    """
    Tests for aa_forum_main_character_alliance_id template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up template

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.template = (
            "{% load aa_forum %}{{ user|aa_forum_main_character_alliance_id }}"
        )

    def test_should_contain_alliance_id_for_users_with_main(self):
        """
        Test should return main character alliance ID for user with a main character

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="3001")

    def test_should_be_dummy_id_for_users_without_main(self):
        """
        Test should return dummy ID (1) for user without main character

        :return:
        :rtype:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")

    def test_should_dummy_id_when_main_is_not_in_an_alliance(self):
        """
        Test should dummy ID (1) when the main character is not in an alliance

        :return:
        :rtype:
        """

        # given
        user = create_fake_user(
            character_id=2012,
            character_name="William Riker",
            corporation_id=2012,
            corporation_name="Starfleet",
            corporation_ticker="SF",
            alliance_id=None,
        )

        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")

    def test_should_be_dummy_id_for_sentinel_user(self):
        """
        Test should return dummy ID (1) for sentinel user

        :return:
        :rtype:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")

    def test_should_be_dummy_id_for_none(self):
        """
        Test should return dummy ID (1) for None

        :return:
        :rtype:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(first=result, second="1")


class TestForumTemplateVariables(TestCase):
    """
    Tests for aa_forum template_variables template tag
    """

    def test_aa_forum_template_variable(self):
        """
        Test aa_forum_template_variable

        :return:
        :rtype:
        """

        rendered = render_template(
            string=(
                "{% load aa_forum %}"
                "{% aa_forum_template_variable foo = content %}"
                "{{ foo }}"
            ),
            context={"content": "bar"},
        )

        self.assertEqual(first=rendered, second="bar")

    def test_should_raise_template_syntax_error(self):
        """
        Test should raise template syntax error

        :return:
        :rtype:
        """

        with self.assertRaisesMessage(
            expected_exception=TemplateSyntaxError,
            expected_message=(
                "'aa_forum_template_variable' tag must be of the form: "
                "{% aa_forum_template_variable <var_name> = <var_value> %}"
            ),
        ):
            render_template(
                string=(
                    "{% load aa_forum %}"
                    "{% aa_forum_template_variable foo %}"
                    "{{ foo }}"
                )
            )

    def test_should_return_personal_message_unread_count_as_empty_string(self):
        """
        Test personal_message_unread_count to return zero

        :return:
        :rtype:
        """

        user = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )

        self.client.force_login(user)

        rendered = render_template(
            string=("{% load aa_forum %}{% personal_message_unread_count 1001 %}")
        )

        self.assertEqual(first=rendered, second="")

    def test_should_return_personal_message_unread_count_as_bootstrap_badge_with_number(
        self,
    ):
        """
        Test personal_message_unread_count to return a bootstrap badge with a number

        :return:
        :rtype:
        """

        # given (creating our personal message)
        user_sender = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )

        user_receiver = create_fake_user(
            character_id=1002,
            character_name="Batman",
            corporation_id=2001,
            corporation_name="Wayne Tech Inc.",
            corporation_ticker="WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )

        PersonalMessage.objects.create(
            sender=user_sender,
            recipient=user_receiver,
            subject="Foo",
            message="Bar",
        )

        # when (template tag is triggered)
        response = personal_message_unread_count(user_receiver)

        # then
        self.assertEqual(
            first=response,
            second=(
                '<span class="badge bg-light aa-forum-badge-personal-messages-unread-count">1</span>'  # pylint: disable=line-too-long
            ),
        )


class TestForumVersionedStatic(TestCase):
    """
    Tests for aa_forum_static template tag
    """

    def test_versioned_static(self):
        """
        Test should return static URL string with a version

        :return:
        :rtype:
        """

        context = {"version": __version__}

        rendered_template = render_template(
            string=(
                "{% load aa_forum %}"
                "{% aa_forum_static 'aa_forum/css/aa-forum.min.css' %}"
            ),
            context=context,
        )

        self.assertEqual(
            first=rendered_template,
            second=f'/static/aa_forum/css/aa-forum.min.css?v={context["version"]}',
        )


class TestHighlightSearchTerm(TestCase):
    """
    Tests for highlight_search_term template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup

        :return:
        :rtype:
        """

        super().setUpClass()

        cls.message_text = "Lorem Ipsum"
        cls.message_with_link = '<a href="https://lorem-ipsum.com">Lorem Ipsum</a>'
        cls.message_with_link_without_href = '<a title="Lorem Ipsum">Lorem Ipsum</a>'
        cls.message_with_link_and_title = (
            '<a href="https://lorem-ipsum.com" title="Lorem Ipsum">Lorem Ipsum</a>'
        )
        cls.message_with_link_and_name = (
            '<a href="https://lorem-ipsum.com" name="Lorem Ipsum">Lorem Ipsum</a>'
        )
        cls.message_with_image = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg"> Lorem Ipsum'
        )
        cls.message_with_image_without_src = '<img alt="Lorem Ipsum"> Lorem Ipsum'
        cls.message_with_image_and_alt = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg" alt="Lorem Ipsum"> '
            "Lorem Ipsum"
        )
        cls.message_with_image_and_title = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg" title="Lorem Ipsum"> '
            "Lorem Ipsum"
        )
        cls.search_term = "Lorem"
        cls.template = (
            "{% load aa_forum %}"
            "{{ message|aa_forum_highlight_search_term:search_term  }}"
        )

    def test_should_highlight_with_just_text(self):
        """
        Test should highlight search term

        :return:
        :rtype:
        """

        context = {"message": self.message_text, "search_term": self.search_term}

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_add_dummy_href(self):
        """
        Test should add a dummy href

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_link_without_href,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<a href="#" title="Lorem Ipsum">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_not_highlight_in_link_href(self):
        """
        Test should not highlight in an a-tag's href attribute

        :return:
        :rtype:
        """

        context = {"message": self.message_with_link, "search_term": self.search_term}

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<a href="https://lorem-ipsum.com">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_not_highlight_in_link_title(self):
        """
        Test should not highlight in an a-tag's title attribute

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_link_and_title,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<a href="https://lorem-ipsum.com" title="Lorem Ipsum">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_not_highlight_in_link_name(self):
        """
        Test should not highlight in an a-tag's name attribute

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_link_and_name,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<a href="https://lorem-ipsum.com" name="Lorem Ipsum">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_not_highlight_in_img_src(self):
        """
        Test should not highlight in an img-tag's src attribute

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_image,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_add_dummy_image_src(self):
        """
        Test should add a dummy image src

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_image_without_src,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<img alt="Lorem Ipsum" src="#"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_not_highlight_in_img_alt(self):
        """
        Test should not highlight in an img-tag's alt attribute

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_image_and_alt,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<img alt="Lorem Ipsum" src="https://lorem-ipsum.com/lorem-ipsum.jpg"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_not_highlight_in_img_title(self):
        """
        Test should not highlight in an img-tag's title attribute

        :return:
        :rtype:
        """

        context = {
            "message": self.message_with_image_and_title,
            "search_term": self.search_term,
        }

        rendered_template = render_template(string=self.template, context=context)

        expected_result = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg" title="Lorem Ipsum"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(first=rendered_template, second=expected_result)


class TestForumDatetime(TestCase):
    """
    Tests for aa_forum_time template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Setup

        :return:
        :rtype:
        """

        super().setUpClass()

        cls.db_datetime = parser.parse(timestr="2021-08-17 05:38:13.887496")
        cls.template = "{% load aa_forum %}{{ message_date|aa_forum_time }}"

    @modify_settings(INSTALLED_APPS={"remove": "timezones"})
    def test_should_return_formatted_date_and_time(self):
        """
        Test should return the formatted date and time without aa-timezones support

        :return:
        :rtype:
        """

        context = {"message_date": self.db_datetime}
        rendered_template = render_template(string=self.template, context=context)
        expected_result = "Aug. 17, 2021, 05:38:13"

        self.assertEqual(first=rendered_template, second=expected_result)

    @modify_settings(INSTALLED_APPS={"append": "timezones"})
    def test_should_return_formatted_date_and_time_with_aa_timezones_support(self):
        """
        Test should return a formatted date and time with aa-timezones support

        :return:
        :rtype:
        """

        context = {"message_date": self.db_datetime}
        rendered_template = render_template(string=self.template, context=context)
        timestamp_from_db_datetime = int(datetime.timestamp(self.db_datetime))
        timezones_url = reverse_url(
            viewname="timezones:index", args=[timestamp_from_db_datetime]
        )
        link_title = "Timezone conversion"
        formatted_forum_date = "Aug. 17, 2021, 05:38:13"

        expected_result = (
            f"{formatted_forum_date} "
            f'<sup>(<a href="{timezones_url}" target="_blank" rel="noopener noreferer" '
            f'title="{link_title}" data-bs-tooltip="aa-forum">'
            '<i class="fa-solid fa-circle-question"></i></a>)</sup>'
        )

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_return_empty_date_and_time_string(self):
        """
        Test should return empty date and time string for message_date = ""

        :return:
        :rtype:
        """

        context = {"message_date": ""}
        rendered_template = render_template(string=self.template, context=context)
        expected_result = ""

        self.assertEqual(first=rendered_template, second=expected_result)

    def test_should_return_empty_date_and_time_string_for_none(self):
        """
        Test should return empty date and time string for message_date = None

        :return:
        :rtype:
        """

        context = {"message_date": None}
        rendered_template = render_template(string=self.template, context=context)
        expected_result = ""

        self.assertEqual(first=rendered_template, second=expected_result)



"""
App settings

This module provides functions to check if certain apps are installed and active
"""

# Django
from django.apps import apps


def discordproxy_installed() -> bool:
    """
    Check if discordproxy is installed and active

    :return:
    :rtype:
    """

    try:
        # Third Party
        from discordproxy.client import (  # pylint: disable=import-outside-toplevel unused-import  # noqa: F401
            DiscordClient,
        )
    except ModuleNotFoundError:
        return False

    return True


def allianceauth_discordbot_installed() -> bool:
    """
    Check if allianceauth-discordbot is installed and active

    :return:
    :rtype:
    """

    return apps.is_installed(app_name="aadiscordbot")


def discord_messaging_proxy_available() -> bool:
    """
    Check if discord messaging proxy is available

    :return:
    :rtype:
    """

    return discordproxy_installed() or allianceauth_discordbot_installed()


def aa_timezones_installed() -> bool:
    """
    Check if aa-timezones is installed and active

    :return:
    :rtype:
    """

    return apps.is_installed(app_name="timezones")
